import os
import sys
import asyncio
from unittest import TestCase, IsolatedAsyncioTestCase
from tempfile import mkdtemp
from shutil import rmtree
from unittest.mock import Mock, patch, mock_open, call, AsyncMock
from jmflashcards.commands import load_config
from jmflashcards.runner import run_command
from jmflashcards.latex import render_latex_to_file
from jmflashcards.parser import FlashCard, Entry, Side, MathSide, ImageSide, \
        TextSide
from jmflashcards.fcdeluxe import FCDFlashCard

CURRENT_DIR = os.path.dirname(os.path.abspath(__file__))
FLASHCARD_DIR = "flashcards"
FLASHCARD_PATH = os.path.join(CURRENT_DIR, FLASHCARD_DIR)
NUM_ENTRIES = 5

def test_run_command():
    async def test():
        await run_command("ls", cwd="/tmp")
    asyncio.run(test())


# TODO Configure the load config procedure
CONFIG_FILE_WORKING ="""
input_dir: /dev/null
output_dir: /dev/urandom
question_keys: 
    - one
    - two
answer_keys:
    - three
    - four
"""

class ConfigurationTestCase(TestCase):
    def test_defaultconfig(self):
        from jmflashcards.commands import INPUT_DIR, OUTPUT_DIR, \
                QUESTION_KEYS, ANSWER_KEYS
        with patch('jmflashcards.commands.open',
                mock_open(read_data="")): 
            result = load_config()
            self.assertIn("input_dir", result)
            self.assertEqual(result['input_dir'], INPUT_DIR)
            self.assertIn("output_dir", result)
            self.assertEqual(result['output_dir'], OUTPUT_DIR)
            self.assertIn("question_keys", result)
            for key in QUESTION_KEYS:
               self.assertIn(key, result['question_keys'])
            self.assertIn("answer_keys", result)
            for key in ANSWER_KEYS:
               self.assertIn(key, result['answer_keys'])

    def test_custom_config(self):
        with patch('jmflashcards.commands.open', 
                mock_open(read_data=CONFIG_FILE_WORKING)): 
            result = load_config()
            self.assertIn("input_dir", result)
            self.assertEqual(result['input_dir'], "/dev/null")
            self.assertIn("output_dir", result)
            self.assertEqual(result['output_dir'], "/dev/urandom")
            self.assertIn("question_keys", result)
            for key in ('one', 'two'):
               self.assertIn(key, result['question_keys'])
            self.assertIn("answer_keys", result)
            for key in ('three', 'four'):
               self.assertIn(key, result['answer_keys'])


class FlashCardTestCase(TestCase):

    def setUp(self):
        self.syncronizer = Mock()
        self.syncronizer.question_keys = ('question',)
        self.syncronizer.answer_keys = ('answer',)
        self.repository = Mock()
        self.repository.syncronizer = self.syncronizer
        self.repository.directory = FLASHCARD_PATH

        self.flashcard = FlashCard('test', self.repository)
        self.flashcard.parse()

    def test_one(self):
        self.assertEqual(len(self.flashcard.entries), NUM_ENTRIES)
        for entry in self.flashcard.entries:
            self.assertTrue(isinstance(entry, Entry))
            self.assertTrue(isinstance(entry.question, Side))
            self.assertTrue(isinstance(entry.answer, Side))

class TextParseTestCase(TestCase):

    def _get_side(self, raw_text, name="question"):
        mock_entry = Mock()
        return Side.build_from_raw(mock_entry, name, raw_text)

    def assert_side_class(self, raw_text, side_class):
        side = self._get_side(raw_text)
        if not isinstance(side, side_class):
            self.fail("Expression '%s' should have been parsed as class '%s' but returned class '%s'" % (raw_text, side_class.__name__, side.__class__.__name__))

    def test_expression_parsing(self):
        self.assert_side_class("texto", TextSide)
        self.assert_side_class("\\$texto", TextSide)
        self.assert_side_class("\\~texto", TextSide)
        self.assert_side_class("\\\\texto", TextSide)
        self.assert_side_class("$texto", MathSide)
        self.assert_side_class("~texto", ImageSide)

    def assert_text_parsing(self, raw_text, expected):
        self.assert_side_class(raw_text, TextSide)
        side = self._get_side(raw_text)
        cured = side.get_cured_text()
        self.assertEqual(expected, cured, 
                "Text '%s' should have rendered like '%s' but rendered '%s'" % ( raw_text, 
                    expected, cured))

    def test_text_parsing(self):
        self.assert_text_parsing("\\$text", "$text")
        self.assert_text_parsing("\\~text", "~text")
        self.assert_text_parsing("\\\\text", "\\text")

from jmflashcards.fcdeluxe import FCDELUXE_DIR_NAME, \
        FCDFlashCardRenderer, FCDRepository, FCDELUXE_HEADER

class FlashCardsDeluxeTestCase(IsolatedAsyncioTestCase):
    def setUp(self):
        self.syncronizer = Mock()
        self.repository = Mock()
        self.repository.syncronizer = self.syncronizer
        self.repository.directory = FLASHCARD_PATH
        self.syncronizer.question_keys = ('question',)
        self.syncronizer.answer_keys = ('answer',)

    async def test_build(self):
        flashcard = FlashCard('test', self.repository)
        flashcard.parse()
        output_dir = mkdtemp(prefix="mockoutput_dir")
        fcd_repository = FCDRepository(output_dir)
        renderer = FCDFlashCardRenderer(fcd_repository)
        fcd_flashcard = await renderer.render(flashcard)

        file_name = flashcard.reference + ".txt"
        file_path = os.path.join(output_dir, FCDELUXE_DIR_NAME, file_name)
        media_dir_path = os.path.join(output_dir, FCDELUXE_DIR_NAME, 
                FCDFlashCard.get_media_dir_name(flashcard.reference))

        self.assertTrue(os.path.exists(file_path))
        self.assertTrue(os.path.isfile(file_path))
        self.assertTrue(os.path.exists(media_dir_path))
        self.assertTrue(os.path.isdir(media_dir_path))

        f = open(file_path)
        text = f.read()
        f.close()

        def fail_msg(msg, text=text):
            line = "."*70
            content = msg, line, text, line
            return "\n".join(content)

        lines = text.split('\n')
        self.assertEqual(lines[0], FCDELUXE_HEADER[:-1])
        line_counter = 0
        for line in lines[1:]:
            if not line:
                continue
            line_counter += 1
            fields = line.split('\t')
            self.assertEqual(len(fields), 6, 
                    fail_msg("Invalid entry fields count", text=repr(line)))
            for f_field in fields[2:]:
                if f_field:
                    file_path = os.path.join(media_dir_path, f_field)
                    self.assertTrue(os.path.exists(file_path), 
                            fail_msg("File '%s' dont exist" % file_path))
        self.assertEqual(line_counter, NUM_ENTRIES)
        rmtree(output_dir)

from jmflashcards.latex import render_latex_to_file, WORK_DIR

class LatexRendererTestCase(IsolatedAsyncioTestCase):
    output_file = "/tmp/output.png"
    tex_expression = "\\alpha"

    async def test_run(self):
        await render_latex_to_file(self.tex_expression, self.output_file)
        self.assertTrue(os.path.exists(self.output_file))

    def tearDown(self):
        if os.path.exists(self.output_file):
            os.remove(self.output_file)

if __name__ == "__main__":
    import unittest
    unittest.main()




import os
import codecs
from traceback import format_exc
from shutil import copyfile
from functools import reduce
import asyncio
import logging
from contextlib import asynccontextmanager, contextmanager
from shutil import copyfile, rmtree
from jmflashcards.errors import JMFCError
from jmflashcards.parser import TextSide, ImageSide, MathSide
from jmflashcards.latex import render_latex_to_file
from jmflashcards.util import mkdir_p, walkdirs

FCDELUXE_HEADER = "Text 1\tText 2\tPicture 1\tPicture 2\tSound 1\tSound 2\n"
FCDELUXE_DIR_NAME = "Flashcards Deluxe"

class FCDFlashCard(object):
    header = FCDELUXE_HEADER
    file_extension = ".txt"

    @staticmethod
    def get_media_dir_name(name):
        return "%s Media" % name

    def __init__(self, reference, repository):
        self.reference = reference
        self.repository = repository
        self.file_name = os.path.basename(self.reference) + self.file_extension
        self.path = os.path.join(self.repository.directory, self.reference + 
                self.file_extension)
        self.media_path = os.path.join(self.repository.directory, 
                FCDFlashCard.get_media_dir_name(self.reference))

    def delete(self):
        os.remove(self.path)
        rmtree(self.media_path)

    def get_date(self):
        filestat = os.stat(self.path)
        return filestat.st_mtime


class FCDFlashCardRenderer(object):
 
    def __init__(self, repository):
        self.repository = repository

    async def render(self, flashcard):
        reference = flashcard.reference
        logging.info("Begin rendering flashcard deluxe: %s" % reference)
        fcd_flashcard = FCDFlashCard(reference, self.repository)
        path = fcd_flashcard.path
        media_path = fcd_flashcard.media_path
        flashcard_dir = os.path.dirname(path)

        # Create directories
        logging.debug("Creating flashcard subdirectory '%s'" % path)
        try:
            mkdir_p(flashcard_dir)
        except:
            msg = "Error creating flashcard directory: '%s'\n%s" % (flashcard_dir, format_exc())
            raise JMFCError(msg)
        if not os.path.exists(media_path):
            logging.debug("Creating flashcard deluxe media directory: %s" % media_path)
            try:
                os.mkdir(media_path)
            except:
                msg = "Error creating flashcard media directory: '%s'" % media_path
                raise JMFCError(msg)
        if not os.path.isdir(media_path):
            msg = "Provided path to build F.C. Deluxe flashcard is not a directory"
            raise JMFCError(msg)

        # Render entries

        logging.debug("Creating flashcard deluxe file: %s" % path)
        lines = []
        for entry in flashcard.entries:
            lines.append(await self.render_entry(entry, media_path))
        f = codecs.open(path, "w", "utf-8")
        f.write(FCDELUXE_HEADER)
        f.writelines(lines)
        f.close()
        return fcd_flashcard

    async def render_entry(self, entry, media_path):
        logging.debug("Building entry: %d" % entry.index)
        sqrenderer = self.get_side_renderer(entry.question, media_path)
        await sqrenderer.render()
        rq = sqrenderer.get_text_tuple()
        sarenderer = self.get_side_renderer(entry.answer, media_path)
        await sqrenderer.render()
        ra = sqrenderer.get_text_tuple()
        res = reduce(lambda x,y: x + list(y), zip(rq, ra), [])
        return "\t".join(res) + "\n"

    def get_side_renderer(self, side, media_path):
        "Given the side object selects the matching renderer object"

        for sc, rc in self.renderer_assignement:
            if issubclass(side.__class__, sc):
                logging.debug("Selecting '%s'" % rc.__name__) 
                return rc(side, media_path)

class SideRenderer(object):
    file_name_template = "entry_%{entry_index}d_%{section_type}s"
    subclasses = []

    def __init__(self, side, media_path):
        self.side = side
        self.entry = side.entry
        self.flashcard = side.flashcard
        self.media_path = media_path
        self.cured_text = side.get_cured_text()

    async def render(self):
        pass

    def get_text_tuple(self):
        return "", "", ""


class TextSideRenderer(SideRenderer):
    def get_text_tuple(self):
        return self.cured_text, "", ""


class FileRenderer(SideRenderer):
    file_name_template = "entry_%(entry_index)d_%(section_side)s.%(extension)s"

    def __init__(self, *args):
        super().__init__(*args)
        values = dict(
                entry_index  = self.entry.index,
                section_side = self.side.name,
                extension = self.get_ext()
                )
        self.dest_file_name = self.file_name_template % values
        self.output_path = os.path.join(self.media_path, self.dest_file_name)

    def get_ext(self):
        return '.png'

    async def render(self):
        return self

    def get_text_tuple(self):
        return "", self.dest_file_name , ""


class ImageSideRenderer(FileRenderer):
    def get_ext(self):
        return os.path.splitext(self.cured_text)

    async def render(self):
        copyfile(self.cured_text, self.output_path)


class MathSideRenderer(FileRenderer):
    async def render(self):
        await render_latex_to_file(self.cured_text, self.output_path)


FCDFlashCardRenderer.renderer_assignement=[(TextSide, TextSideRenderer),
                                           (ImageSide, ImageSideRenderer),
                                           (MathSide, MathSideRenderer)]


class FCDRepository(object):
    dir_name = FCDELUXE_DIR_NAME 
    flashcard_class = FCDFlashCard
    renderer_class = FCDFlashCardRenderer

    def __init__(self, dropbox_dir):
        self.dropbox_dir = dropbox_dir
        self.directory = os.path.join(dropbox_dir, self.dir_name)
        self.renderer = self.renderer_class(self)

        if not os.path.exists(self.dropbox_dir):
            raise JMFCError("Output directory dont exist")
        if not os.path.isdir(self.dropbox_dir):
            raise JMFCError("Output path is not a directory")
        if not os.path.exists(self.directory):
            try:
                os.mkdir(self.directory)
            except OSError:
                raise JMFCError("Unable to create Flash Cards Deluxe directory")
        if not os.path.isdir(self.directory):
            raise JMFCError("Flash Cards deluxe path is not a directory")

    def iter_references(self):
        for dirpath, dirnames, filenames in walkdirs(self.directory):
            # ignore media directories
            if " Media" in dirpath and dirpath[-6:] == " Media":
                continue
            for filename in filenames:
                name, ext = os.path.splitext(filename)
                if ext == ".txt" and (name + " Media") in dirnames:
                    yield os.path.join(dirpath, name)

    def references(self):
        return list(self.iter_references())

    def iter_flashcards(self):
        for ref in self.iter_references():
            yield self[ref]

    def flashcards(self):
        return list(self.iter_flashcards())

    def __getitem__(self, name):
        for ref in self.iter_references():
            if ref == name:
                return self.flashcard_class(ref, self)
        raise KeyError(name)

    def __iter__(self):
        for ref in self.iter_references():
            yield ref, self[ref]



from traceback import format_exception
class JMFCError(Exception): pass

# Run comand error
class JMFCCommandError(JMFCError): 
    def __init__(self, command, path, rc, stdout, stderr):
        self.command = command
        self.path = path
        self.rc = rc
        self.stdout = stdout
        self.stderr = stderr

    def __str__(self):
        res = ["Error %d running command: '%s'" % (self.rc, self.command)]
        if not self.path is None:
            res.append("On path '%s'" % self.path)
        if self.stdout:
            res.append("Returned on stdout:\n%s" % self.stdout)
        if self.stderr:
            res.append("Returned on stderr:\n%s" % self.stderr)
        return "\n".join(res)

# Parsing errors

class JMFCEParsingError(JMFCError):
    pass

class JMFCParsingEntriesError(JMFCEParsingError):
    divider_symbol = "."
    divider_symbol_count = 70
    divider_line = "\n" + divider_symbol_count * divider_symbol + "\n"

    def __init__(self, flashcard, errors):
        self.flashcard = flashcard
        self.errors = errors

    def _format_exception(self, e_type, e_value, traceback):
        if issubclass(e_type, JMFCError):
            return str(e_value)
        else:
            return "\n".join(format_exception(e_type, e_value, traceback))

    def __str__(self):
        reference = self.flashcard.reference
        header = "Error parsing flashcard '%s' entries:" % reference
        error_msgs = [ self._format_exception(*e) for e in self.errors ]
        error_msgs.insert(0, header)
        return self.divider_line.join(error_msgs)

class JMFCEntryError(JMFCEParsingError):
    header_template = "Error handling entry %d:"

    def __init__(self, entry, msg):
        self.entry = entry
        self.msg = msg

    def _get_header_template_args(self):
        return (self.entry.index, )

    def _get_header(self):
        return self.header_template % self._get_header_template_args()

    def __str__(self):
        header = self._get_header()
        return "%s\n%s" % (header, self.msg)

class JMFCEntrySideError(JMFCEntryError):
    header_template = "Error handling entry %d %s '%s':"

    def __init__(self, side, msg):
        self.side = side
        self.msg = msg

    def _get_header_template_args(self):
        return (self.side.entry.index, self.side.side, self.side.raw_text)


# Flashcards deluxe error
class FCDError(JMFCError): pass



import os
import codecs
from contextlib import asynccontextmanager
import logging
from jmflashcards.runner import run_command
from jmflashcards.util import get_random_name

LATEX_EXTENSION = ".tex"
DVI_EXTENSION = ".dvi"
PNG_EXTENSION = ".png"
LATEX_BUILD_EXTENSIONS = ".log", ".aux", LATEX_EXTENSION
DVI_TEMPLATE='dvipng -T tight -x 1200 -z 9 -bg transparent -o "%s" "%s"'
WORK_DIR = '/tmp'
LATEX_TEMPLATE = r'''
\documentclass{article} 
\usepackage{amsmath}
\usepackage{amsthm}
\usepackage{amssymb}
\usepackage{bm}
\newcommand{\mx}[1]{\mathbf{\bm{#1}}} %% Matrix command
\newcommand{\vc}[1]{\mathbf{\bm{#1}}} %% Vector command 
\newcommand{\T}{\text{T}}                %% Transpose
\pagestyle{empty} 
\begin{document} 
$%s$
\end{document}
'''

async def render_latex_to_file(expression, output_file_path):
    logging.info("Rendering latex expression: '%s'" % expression)
    name = get_random_name()

    #TeX template render
    tex_file_content = LATEX_TEMPLATE % expression
    tex_file_path = os.path.join(WORK_DIR, name + LATEX_EXTENSION)
    with codecs.open(tex_file_path, "w", "utf-8") as f:
        f.write(tex_file_content)

    #Compile TeX and remove superfluous files
    await run_command("latex %s" % tex_file_path, cwd=WORK_DIR)
    for ext in LATEX_BUILD_EXTENSIONS:
        os.remove(os.path.join(WORK_DIR, name + ext)) 


    #Convert dvi to png and take care of the last evidences
    dvi_path = os.path.join(WORK_DIR, name + DVI_EXTENSION)
    await run_command(DVI_TEMPLATE % (output_file_path, dvi_path), cwd=WORK_DIR)
    os.remove(dvi_path)




import asyncio
import logging
from jmflashcards.errors import JMFCError
from jmflashcards.parser import Repository 
from jmflashcards.fcdeluxe import FCDRepository 

class Syncronizer(object):
    fcd_repository_class = FCDRepository
    repository_class = Repository

    def __init__(self, output_dir,  directory, question_keys, answer_keys,
            empty=True):
        self.output_dir = output_dir
        self.directory = directory
        self.question_keys = question_keys
        self.answer_keys = answer_keys
        self.repository = self.repository_class(directory, self)
        self.fcd_repository = self.fcd_repository_class(output_dir)
        self.empty = empty

    def get_flashcards_status(self):
        fcd_references = set(self.fcd_repository.references())
        logging.debug("fcd_references: %s" % ", ".join(fcd_references))
        references = set(self.repository.references())
        logging.debug("orig_references: %s" % ", ".join(references))
        to_delete = fcd_references.difference(references)
        logging.debug("delete_references: %s" % ", ".join(to_delete))
        new = references.difference(fcd_references)
        logging.debug("new_references: %s" % ", ".join(new))
        common = fcd_references.intersection(references)
        updated = [ x for x in self._iter_updated(common) ]
        logging.debug("Sync running to update:%s    to create:%s    to delete:%s" % ("\n".join(updated), "\n".join(new), "\n".join(to_delete)))
        return updated, list(new), list(to_delete)

    def _iter_updated(self, common):
        for ref in common:
            flashcard = self.repository[ref]
            fcd_flashcard = self.fcd_repository[ref]
            if fcd_flashcard.get_date() < flashcard.get_date():
                yield ref

    def sync(self):
        updated, new, to_delete = self.get_flashcards_status()
        if updated or new:
            print("Starting syncronization")
            if new: print ("New: %d" % len(new))
            if updated: print ("Updating: %d" % len(updated))
            asyncio.run(self.sync2(new+updated))
        elif to_delete:
            print ("Removing: %d" % len(to_delete))
            for fc in to_delete:
                self.delete_flashcard(fc)
        else:
            print("New: %d" % len(new))
            print("Updating: %d" % len(updated))
            print("Removing: %d" % len(to_delete))
            print("Nothing to be done")

    async def sync2(self, coroutines):
        tasks = []
        for fc in coroutines:
            task = asyncio.create_task(self.render_flashcard(fc))
            tasks.append(task)
        await asyncio.gather(*tasks)

    async def render_flashcard(self, ref):
        logging.info("Building flashcard '%s'" % ref)
        if self.empty:
            return
        flashcard = self.repository[ref]
        try:
            flashcard.parse()
            await self.fcd_repository.renderer.render(flashcard)
        except JMFCError as e:
            logging.error(str(e))

    async def delete_flashcard(self, ref):
        logging.info("Removing flashcard '%s'" % ref)
        if not self.empty:
            fcd_flashcard = self.fcd_repository[ref]
            fcd_flashcard.delete()












          



import os
from argparse import ArgumentParser
import coloredlogs
import yaml

USAGE = "%prog [options] <flashcard dir>"
LOGGING_FORMAT = "[%(levelname)s] %(message)s"
INPUT_DIR = "~/Dropbox/flashcards"
OUTPUT_DIR = "~/Dropbox"
CONFIG_FILE_PATH = '~/.config/jmflashcards/config.yaml'
QUESTION_KEYS = ("question","pregunta")
ANSWER_KEYS = ("answer","respuesta")

def get_logging_level(verbosity):
    if verbosity == 1:
        return 'WARNING'
    elif verbosity == 2:
        return 'INFO'
    elif verbosity >= 3:
        return 'DEBUG'
    else:
        return 'ERROR'

def init_logging(verbosity):
    coloredlogs.install(fmt=LOGGING_FORMAT, 
            level=get_logging_level(verbosity))

def load_config():
    try:
        with open(CONFIG_FILE_PATH, 'r') as f:
            txt = f.read()
            try:
                data = yaml.load(txt, Loader=yaml.Loader)
            except yaml.yamlError as exc:
                if hasattr(exc, 'problem_mark'):
                    mark = exc.problem_mark
                    print("Config error in position (%s:%s)" % (mark.line+1,
                            mark.column+1))
                else:
                    print("Config error")
                exit(1)
            data = {} if data is None else data
    except:
        data = {}
    result = {}
    result['input_dir'] = data.get('input_dir', INPUT_DIR)
    result['output_dir'] = data.get('output_dir', OUTPUT_DIR)
    result['question_keys'] = data.get('question_keys', QUESTION_KEYS)
    result['answer_keys'] = data.get('answer_keys', ANSWER_KEYS)
    return result

def get_argument_parser(config):
    parser = ArgumentParser()
    parser.add_argument("-i", "--input-dir",  
            dest="input_dir", default=config['input_dir'], 
            help="where to find flashcards definitions")
    parser.add_argument("-o", "--output-dir",  
             dest="output_dir", default=config['output_dir'], 
            help="path to the output directory")
    parser.add_argument("-V" , action="count", dest="verbosity", default=0,
            help="sets verbosity level")
    parser.add_argument("-e" , dest="empty", default=False, action="store_true",
            help="doesnt apply actions")
    parser.add_argument("-q" , "--question_key", action="append",
            dest="question_keys", help="add key for questions",
            default=config['question_keys'])
    parser.add_argument("-a" , "--answer_key", action="append",
            dest="answer_keys", help="add key for answer",
            default=config['answer_keys'])
    return parser

def initialize():
    config = load_config()
    parser = get_argument_parser(config)
    args = parser.parse_args()
    init_logging(args.verbosity)
    return args

def run_syncronize():
    from jmflashcards.syncronizer import Syncronizer
    args = initialize()
    output_dir = os.path.expanduser(args.output_dir)
    directory = os.path.expanduser(args.input_dir)
    question_keys = args.question_keys
    answer_keys = args.answer_keys
    syncronizer = Syncronizer(output_dir, directory, question_keys, 
            answer_keys, empty=args.empty)
    syncronizer.sync()



import os
from shutil import copyfile
from contextlib import contextmanager
from jmflashcards.error import JMFCError
from latex import render_to_file

PREGUNTA = 1
RESPUESTA = 2

class CompilerError(JMFCError): 
    def __init__(self, compiler, msg):
        self.compiler = compiler
        self.msg = msg

    def __str__(self):
        return "Error compiling %s of item %d:\n%s" % (self.compiler.p_o_r,
                self.compiler.numero, self.msg)

# p_o_r : pregunta o respuesta

class BaseExpressionCompiler(object):
    def __init__(self, definition_dir, definition, numero, p_o_r, destdir):
        self.definition_dir = definition_dir
        self.definition = definition
        self.numero = numero
        self.p_o_r = p_o_r
        self.destdir = destdir

    def _compile(self):
        pass

    def _get_text(self):
        return ""

    def _get_image_file_name(self):
        return ""

    def _get_audio_file_name(self):
        return ""

    def _error(self, msg):
        raise CompilerError(self, msg)

    def compile(self):
        self._compile()
        return (self.get_text(), self._get_image_file_name(), 
            self._get_audio_file_name())


class ExpressionToFileCompiler(BaseExpressionCompiler):
    extension = None

    def __init__(self, definition_dir, definition, numero, p_o_r, destdir, 
            overwrite=True):
        super(ExpressionToFileCompiler, self).__init__(definition_dir, 
                definition, numero, p_o_r, destdir)
        self.overwrite = overwrite

    @contextmanager
    def _file_compiled(self):
        raise NotImplementedError()

    def _compile(self):
        with self._file_compiled() as file_path:
            self._put_file_in_place(file_path)

    def _get_file_name(self):
        return "%s_%d" % (self.p_o_r, self.numero)

    def _put_file_in_place(self, file_path):
        self._copy_and_rename_file(file_path, self.get_file_name(),  
                extension=self.extension, overwrite=self.overwrite)

    def _copy_and_rename_file(self, original_file, name, extension=None, 
            overwrite=True):
        if not os.path.exist(original_file):
            self._error("File doesnt exist '%s'" % original_file)
        if not os.path.isfile(original_file):
            self._error("Path '%s' is not file" % original_file)
        if not os.path.exist(self.destdir):
            self._error("destdir doesnt exist '%s'" % original_file)
        if not os.path.isdir(original_file):
            self._error("destdir '%s' is not dir" % original_file)
        if not os.path.exist(self.destdir):
            self._error("destdir doesnt exist '%s'" % original_file)
        if extension is None:
            extension = os.path.splitext(original_file)[1]
        dest_path = os.path.join(self.destdir, name + extension)
        if os.path.exist(dest_path) and not overwrite:
            raise CompilerError("compiled file already exist '%s'" % dest_path)
        try:
            copyfile(original_file, dest_path)
        except:
            self._error("Error copying file '%s' to '%s'" % (original_file,
                dest_path))

class ImageCompiler(ExpressionToFileCompiler):
    @contextmanager
    def _file_compiled(self):
        yield os.path.join(self.definition_dir, self.definition)

class LatexCompiler(ExpressionToFileCompiler):
    latex_renderer = render_to_file

    def __init__(self, definition_dir, definition, numero, p_o_r, destdir, 
            overwrite=True, workdir="/tmp"):
        super(ImageCompiler, self).__init__(definition_dir, definition, numero, 
                p_o_r, destdir, overwrite=overwrite)
        self.workdir = workdir

    def _get_renderer_arguments(self):
        return (self.definition,) , {"workdir": self.workdir}

    @contextmanager
    def _file_compiled(self):
        args, kwargs = self._get_renderer_arguments()
        with self.latex_renderer(*args, **kwargs) as pngformula_path:
            yield pngformula_path





import os
import sys
import codecs
import yaml
import logging
from jmflashcards.errors import JMFCError, JMFCParsingEntriesError, \
        JMFCEntryError
from jmflashcards.util import walkdirs

MATH_SECTION_SYMBOL = "$"
IMAGE_SECTION_SYMBOL = "~"
ESCAPE_SYMBOL = "\\"

FLASHCARD_FILE_NAME = "flashcard.yaml"

class FlashCard(object):
    flashcard_file_name = FLASHCARD_FILE_NAME
    entry_class = None

    def __init__(self, reference, repository):
        # TODO Work with relative directories
        self.repository = repository
        self.directory = repository.directory
        self.reference = reference
        self.definition_path = os.path.join(repository.directory, reference, 
                                            self.flashcard_file_name)  
        self.entries = []
        self.parsed = False

    def parse(self):
        logging.info("Parsing flashcard: %s" % self.reference)
        try:
            with codecs.open(self.definition_path, "r", "utf-8") as f:
                self.raw_entries = yaml.load(f, Loader=yaml.Loader)
        except IOError:
            raise JMFCError("Unable to read flashcard file: %s" % self.definition_path)
        except yaml.YAMLError as e:
            raise JMFCError("Unable to parse yaml file:\n%s" % str(e))
        except UnicodeDecodeError as e:
            raise JMFCError("Unable to decode UTF-8:\n%s" % str(e))

        if not isinstance(self.raw_entries, list):
            raise JMFCError("Flashcard entries must be a list")

        errors = []
        for i, rentry in enumerate(self.raw_entries, start=1):
            try:
                entry = self.entry_class(self, rentry, i)
                entry.parse()
            except:
                errors.append(sys.exc_info())
                continue
            self.entries.append(entry)
        if errors:
            raise JMFCParsingEntriesError(self, errors)
        self.parsed = True

    def get_date(self):
        stat = os.stat(self.definition_path)
        return stat.st_mtime


class Entry(object):
    boolean_yes = "si"
    boolean_no = "no"

    side_class = None
    
    def __init__(self, flashcard, raw_entry, index):
        logging.debug("Parsing entry: %s" % index)
        self.index = index
        self.flashcard = flashcard
        self.reference = flashcard.reference
        self.repository = flashcard.repository
        self.syncronizer = self.repository.syncronizer
        self.question_keys = self.syncronizer.question_keys
        self.answer_keys = self.syncronizer.answer_keys
        self.raw_entry = raw_entry
        if not isinstance(raw_entry, dict):
            raise JMFCEntryError(self, "Raw entry must be a dictionary")

    def parse(self):
        self.raw_question, self.raw_answer = self._parse_raw_entry(self.raw_entry)
        self.question = self._get_side(self.raw_question, "question")
        self.answer = self._get_side(self.raw_answer, "answer")

    def _parse_raw_side(self, raw_side):
        if not isinstance(raw_side, str):
            if isinstance(raw_side, bool):
                raw_side = self.boolean_yes if raw_side else self.boolean_no
            elif raw_side is None:
                logging.warning("Empty side in flashcard '%s' entry %d" % (self.reference, self.index))
                raw_side = " "
            else:
                raise JMFCEntryError(self, "Invalid raw side '%s'" % str(raw_side))
        if "\n" in raw_side: 
            logging.debug("New line character in entry, replacing it with space")
            raw_side = raw_side.replace("\n", " ")
        return raw_side

    def _parse_raw_entry(self, raw_entry):
        for key in raw_entry:
            if key.lower() in self.question_keys:
                raw_question = self._parse_raw_side(raw_entry[key])
                break
        else:
            raise JMFCEntryError(self, "Entry dont have a valid question")
        for key in raw_entry:
            if key.lower() in self.answer_keys:
                raw_answer = self._parse_raw_side(raw_entry[key])
                break
        else:
            raise JMFCEntryError(self, "Entry dont have a valid answer")
        return raw_question, raw_answer

    def _get_side(self, raw_text, side_name):
        return self.side_class.build_from_raw(self, side_name, raw_text)

FlashCard.entry_class = Entry

class Side(object):
    side_types = []
    type_name = ''

    @classmethod
    def build_from_raw(cls, entry, name, raw_text):
        for st in cls.side_types:
            if st.check_raw_text(raw_text):
                logging.debug("Setting '%s' as side class '%s'" % (raw_text, st.__name__))
                return st(entry, name, raw_text)
        else:
            raise JMFCEntryError(entry, "Unable to assign side type to '%s'" % raw_text)

    def __init__(self, entry, name, raw_text):
        logging.debug("Parsing %s text: '%s'" % (name, raw_text))
        self.entry = entry
        self.flashcard = entry.flashcard
        self.name = name
        self.raw_text = raw_text

Entry.side_class = Side

class DelimiterSideMixin(object):
    @classmethod
    def check_raw_text(cls, raw_text):
        if raw_text[0] == cls.delimiter_symbol:
            return True
        return False

    def get_cured_text(self):
        return self.raw_text[1:]

class MathSide(DelimiterSideMixin, Side):
    delimiter_symbol = MATH_SECTION_SYMBOL
    type_name = 'MATH'

class ImageSide(DelimiterSideMixin, Side):
    delimiter_symbol = IMAGE_SECTION_SYMBOL 
    type_name = 'IMAGE'

class TextSide(Side):
    type_name = 'TEXT'
    escaped_symbols = MATH_SECTION_SYMBOL, IMAGE_SECTION_SYMBOL, ESCAPE_SYMBOL

    @classmethod
    def check_raw_text(cls, raw_section):
        return True

    def get_cured_text(self):
        text = self.raw_text
        if len(text) < 2:
            return text
        first = text[0]
        second = text[1]
        if first == ESCAPE_SYMBOL:
            if second in self.escaped_symbols or second == ESCAPE_SYMBOL:
                    text = text[1:]
        return text

Side.side_types = (MathSide, ImageSide, TextSide)

class Repository(object):
    flashcard_class = FlashCard
    flashcard_file_name = FLASHCARD_FILE_NAME

    def __init__(self, directory, syncronizer):
        logging.debug("Initializing flashcard repository on: %s" % directory)
        self.directory = directory
        self.syncronizer = syncronizer

    def walk_flashcards(self):
        for dirpath, dirnames, filenames in walkdirs(self.directory):
            if not dirpath or dirnames:
                continue
            if self.flashcard_file_name in filenames:
                yield dirpath, dirnames, filenames

    def iter_references(self):
        for dirpath, dirnames, filenames in self.walk_flashcards():
            yield dirpath

    def references(self):
        return list(self.iter_references())

    def iter_flashcards(self):
        for ref in self.iter_references():
            yield self[ref]

    def flashcards(self):
        return list(self.iter_flashcards())

    def parse(self):
        for f in self.iter_flashcards():
            f.parse()

    def __getitem__(self, reference):
        for dirpath, dirnames, filenames in self.walk_flashcards():
            if reference == dirpath:
                return self.flashcard_class(reference, self)
        raise KeyError(reference)

    def __iter__(self):
        for ref in self.iter_references():
            yield ref, self[ref]






import os
import errno
import random
from string import ascii_letters, digits

chars = ascii_letters + digits

def get_random_name(nchars=30):
    return ''.join([ random.choice(chars) for x in range(0, nchars) ])
    
def mkdir_p(path):
    try:
        os.makedirs(path)
    except OSError as exc:
        if exc.errno == errno.EEXIST and os.path.isdir(path):
            pass
        else:
            raise

def walkdirs(path):
    n = len(path)
    for dirpath, dirnames, filenames in os.walk(path):
        dirpath = dirpath[n+1:]
        yield dirpath, dirnames, filenames




import asyncio
from asyncio.subprocess import PIPE
import shlex

from .errors import JMFCCommandError

async def _run_command(args, cwd=None):
    args = shlex.split(args)
    cmd = args[0]
    args = args[1:]
    p = await asyncio.create_subprocess_exec(cmd, *args, stdout=PIPE, stderr=PIPE,
            cwd=cwd)
    stdout, stderr = await p.communicate()
    rc = p.returncode
    stdout = stdout.decode()
    stderr = stderr.decode()
    return rc, stdout, stderr

async def run_command(args, cwd=None):
    rc, stdout, stderr = await _run_command(args, cwd=cwd)
    if rc != 0:
        raise JMFCCommandError(args, cwd, rc, stdout, stderr)




import sys
from logging import getLogger, StreamHandler, DEBUG, INFO, WARNING, ERROR, CRITICAL

def get_logger(level=WARNING):
    logger = getLogger()
    logger.setLevel(level)
    logger.setHandler(StreamHandler(stream=sys.stderr))
    return logger

